-- Creates a scheduler class, for "cron" and "at"-like functionality
-- Going to handle "at" first.  But the basic idea holds

local MAJOR_VERSION = "YakScheduler-2.0"
local MINOR_VERSION = 0

if not AceLibrary then error(MAJOR_VERSION .. " requires AceLibrary.") end
if not AceLibrary:IsNewVersion(MAJOR_VERSION, MINOR_VERSION) then return end

local YakHelpers = AceLibrary("YakHelpers-2.0")

local YakScheduler = {}

function YakScheduler:Create()
   result = {at_queue = {}, cron_queue = {}, last_purged = 0, log = YakHelpers.YAK_LOGGER:Create(), last_check = 0, granularity = 0.25, now = 0}
   setmetatable(result, self)
   self.__index = self
--   result.log:SetLog(true)
   return result
end

function YakScheduler:SetGranularity( new_value )
   self.granularity = new_value
end

function YakScheduler:OnUpdate()
   self.now = GetTime()
   if (self.now - self.last_check) > self.granularity then
      if table.getn(self.at_queue) > 0 then
	 self:UpdateAtQueue(self.now)
      end
      self:UpdateCronQueue(self.now)
      self.last_check = self.now
   end
end

function YakScheduler:UpdateAtQueue( now )
   -- the "at queue" is a table holding {at_time, id, closure, done} tuples indexed
   -- by integer
   for k,v in ipairs(self.at_queue) do
      if now > v[1] and (v[4] == false) then
--         self.log:Log("Doing at-job " .. v[2])
         v[3]()
	 v[4] = true
      end
   end
   if (now > (self.last_purged + 60)) or (table.getn(self.at_queue) > 60) then
      self:PurgeDoneJobs()
      self.last_purged = now
   end
end

function YakScheduler:PurgeDoneJobs()
   local tasks_to_keep = {}
   for k,v in ipairs( self.at_queue ) do
      if v[4] == false then
	 table.insert(tasks_to_keep,v)
      else
--	 self.log:Log("Purging task "..v[2])
      end
   end
   self.at_queue = tasks_to_keep
end

function YakScheduler:UpdateCronQueue( now )
   for k,v in pairs(self.cron_queue) do
      if v[2] < now then
--	 self.log:Log("Doing cron-job " .. k )
	 v[3]()
	 v[2] = now + v[1]
      end
   end
end

function YakScheduler:At(time, id, closure)
   table.insert( self.at_queue, {time, id, closure, false} )
end

function YakScheduler:In( dt, id, closure )
--   self.log:Log("Time" .. GetTime().."; Adding at-job " .. id.. " at "..GetTime()+dt)
--   self.log:Log(table.getn(self.at_queue) .. "job in queue")
   table.insert( self.at_queue, {GetTime()+dt, id, closure, false} )
end

function YakScheduler:Every( dt, id, closure )
   self.cron_queue[id] = {dt, GetTime()+dt, closure}
end

function YakScheduler:SetCronDelay(id, new_dt)
   self.cron_queue[id][1] = new_dt
end

function YakScheduler:RemoveCronJob(id)
   table.remove(self.cron_queue, id)
end

local function activate( self, oldlib, olddeactivate )
   if oldlib then
      self.Scheduler = oldlib.Scheduler
      self.Frame = oldlib.Frame
   end

   if not self.scheduler then
      self.Scheduler = self:Create()
   end

   if not self.Frame then
      self.Frame = CreateFrame("Frame")
--      self.Frame:Hide()
      self.Frame.name = "YakScheduler-2.0 Frame"
      self.Frame:SetScript("OnUpdate", function() self.Scheduler:OnUpdate(elapsed) end)
      self.Frame.owner = self
   end

   if olddeactivate then olddeactivate(oldlib) end

end

AceLibrary:Register(YakScheduler, MAJOR_VERSION, MINOR_VERSION, activate )
YakScheduler = nil


